#!/usr/bin/env python2.6 

__author__="enrico"
__date__ ="$14.05.2010 14:29:21$"

import sys 
if sys.platform == "win32":  
    sys.path.append("i:\panda3d")


from panda3d.core import TextNode 
from panda3d.core import Point2,Point3,Vec3,Vec4
from direct.gui.OnscreenText import OnscreenText
from direct.task.Task import Task
from math import sin, cos, asin, acos, pi
import cPickle, sys, os
from panda3d.core import loadPrcFileData
from panda3d.core import Filename
from pandac.PandaModules import NodePath
from direct.showbase.ShowBase import ShowBase
from pandac.PandaModules import CollisionTraverser, CollisionHandlerPusher
from pandac.PandaModules import CollisionNode, CollisionSphere, BitMask32


DegreeToRadiant = pi / 180.0
RadiantToDegree = 180.0 / pi
BasePath_lvl =os.path.dirname(os.path.abspath(sys.argv[0])) + "/.."
BasePath = Filename.fromOsSpecific(os.path.dirname(os.path.abspath(sys.argv[0])) + "/..").getFullpath()
ModelPath = BasePath + "/models/"
TexturePath = BasePath + "/textures/"
MapPath = BasePath_lvl + "/maps/"


BackgroundHeight = -10.0
LevelTileHeight = 0.0
PlayerHeight = 0.1
EnemyHeight = 0.1
CameraHeight = 20.0

TurnRate = 360.0 # degrees to strafe in one second
Acceleration = 10.0 # ship acceleration in units / (sec*sec)
MaxVelocity = 6.0
MaxVelocitySquared = MaxVelocity ** 2

BulletLife = 5     #How long bullets stay on screen before removed
BulletRepeat = .2  #How often bullets can be fired
BulletSpeed = 10   #Speed bullets move

def myNormalize(myVec):
    myVec.normalize()
    return myVec

def restrain(i, mn = -1, mx = 1): 
    "A simple function to make sure a value is in a given range, -1 to 1 by default"
    return min(max(i, mn), mx)

def genLabelText(text, i):
  return OnscreenText(text = text, pos = (-1.3, .95-.05*i), fg=(1,1,0,1),
                      align = TextNode.ALeft, scale = .05)



def loadObject(model, texture, parent, position = Point2(0,0), depth = LevelTileHeight, scale = 1, 
                color=(1,1,1,1), transparency = True):
    object = loader.loadModel(ModelPath + model)
    if parent != None:
        object.reparentTo(parent)

    object.setPos(Point3(position.getX(), position.getY(), depth))
    #print "do not rotate, instead export model correctly" # disable for release 
    object.setHpr(0.0, 90, 0)

    object.setScale(scale)

    object.setBin("unsorted", 0)

    object.setDepthTest(True)

    if transparency:
        object.setTransparency(1)

    if texture: 
        texture = loader.loadTexture(TexturePath + texture + ".png")
        object.setTexture(texture, 1)

    if len(color) == 3:
        object.composeColorScale(color[0], color[1], color[2], 1.0)
    else:
        object.composeColorScale(color[0], color[1], color[2], color[3])
    return object

class Game(ShowBase):
    def __init__(self):
        ShowBase.__init__(self)
        base.disableMouse()

        genLabelText("ESC: Quit", 0)
        genLabelText("d: Accelerate", 1)
        genLabelText("s: Break", 2)
        genLabelText("[Space Bar]: Fire", 3)
        genLabelText("Mouse: Aim", 4)


        self.enemyCounterText = genLabelText("Enemies: ", 10)

        # prepare collision detection
        self.collisionTraverser = CollisionTraverser()
        self.collisionHandler = CollisionHandlerPusher()
        base.cTrav = self.collisionTraverser
        base.cTrav.setRespectPrevTransform(True)

        self.ship = render.attachNewNode("ship")
        self.ship.setTwoSided(True)
        self.shipHeading = Vec3(0, 0, 0)
        shipObject = loadObject("plane", "ship2", self.ship, depth=PlayerHeight, transparency=True)
        self.setVelocity(self.ship, Vec3(0, 0, 0))
        shipCollider = self.ship.attachNewNode(CollisionNode("player ship"))
        shipCollider.node().addSolid(CollisionSphere(0,0,0, 0.5))
        #shipCollider.show()
        shipCollider.node().setFromCollideMask(BitMask32(0x1))
        shipCollider.node().setIntoCollideMask(BitMask32.allOff())
        self.collisionHandler.addCollider(shipCollider, self.ship)
        base.cTrav.addCollider(shipCollider, self.collisionHandler)


        self.gameTask = taskMgr.add(self.gameLoop, "game loop")
        self.gameTask.last = 0

        self.cameraTask = taskMgr.add(self.positionCamera, "position camera")

        # For a traverser to actually do collisions, you need to call
        # traverser.traverse() on a part of the scene. Fortunatly, base has a
        # task that does this for the entire scene once a frame. This sets up our
        # traverser as the one to be called automatically
        base.cTrav = self.collisionTraverser
        

        self.levelRootNode = render.attachNewNode("level")
        self.levelRootNode.setTwoSided(True)
        dimensions, self.tiles = self.loadLevel(MapPath + "testlevel/first.csv")

        self.walls = [ Point3(position[0], position[1], LevelTileHeight)  for position, token in self.tiles.iteritems() if token == "X" ]

        startPositions = [ position for position, token in self.tiles.iteritems() if token == "s" or token == "S" ]
        oldPosition = self.ship.getPos()
        oldPosition[0] = startPositions[0][0]
        oldPosition[2] = PlayerHeight
        oldPosition[1] = startPositions[0][1]
        self.ship.setPos(oldPosition)

        print "Ship should be at " + str(oldPosition)
        
        self.loadEnemies([ (position, token) for position, token in self.tiles.iteritems() if token == "1" or token == "2"])
        self.initialNumberOfEnemies = len(self.enemies)
        self.enemyCounterText.setText("Enemies: %i/%i" % (self.initialNumberOfEnemies, self.initialNumberOfEnemies))

        self.prepareInput()

        self.mouseInputTask = taskMgr.add(self.mouseMovement, "mouse movement")


        self.gameTask.nextBullet = 0
        self.bullets = []


        self.neededTimeForLevel = 0.0
    
    def loadEnemies(self, enemyTiles):
        self.enemies = []
        
        for position, token in enemyTiles:
            if token == "1" or token == "2":
                enemy = render.attachNewNode("enemy1_" + str(len(self.enemies)+1))
                enemy.setTwoSided(True)
                enemyObject = loadObject("enemy1", "enemy1", enemy, depth=EnemyHeight, transparency=True)
                self.setVelocity(enemy, Vec3(0, 0, 0))
                oldPosition = enemy.getPos()
                oldPosition[0] = position[0]
                oldPosition[2] = EnemyHeight
                oldPosition[1] = position[1]
                enemy.setPos(oldPosition)
                
                print "enemy1_" + str(len(self.enemies)+1) + " should be at " + str(oldPosition)
                self.enemies.append(enemy)

    

    def setKey(self, key, value):
        self.keys[key] = value

    def prepareInput(self):
        self.keys = {"strafe left":False, "strafe right":False, 
                    "accelerate":False, "break":False,
                    "fire":False
                    }


        self.accept("escape", sys.exit)

        self.accept("a", self.setKey, ["strafe left", True])
        self.accept("a-up", self.setKey, ["strafe left", False])

        self.accept("d", self.setKey, ["strafe right", True])
        self.accept("d-up", self.setKey, ["strafe right", False])

        self.accept("w", self.setKey, ["accelerate", True])
        self.accept("w-up", self.setKey, ["accelerate", False])

        self.accept("s", self.setKey, ["break", True])
        self.accept("s-up", self.setKey, ["break", False])

        self.accept("space", self.setKey, ["fire", True])
        self.accept("space-up", self.setKey, ["fire", False])

        self.accept("o", base.oobe)

    def setVelocity(self, object, value):
        list = [value[0], value[1], value[2]]
        object.setTag("velocity", cPickle.dumps(list))

    def getVelocity(self, object):
        list = cPickle.loads(object.getTag("velocity"))
        return Vec3(list[0], list[1], list[2])

    def updatePosition(self, object, dt):
        velocity = self.getVelocity(object)
        newPosition = object.getPos() + (velocity*dt)

        object.setFluidPos(newPosition)

    def setExpires(self, object, value):
        object.setTag("expires", str(value))

    def getExpires(self, object):
        return float(object.getTag("expires"))

    def updateShip(self, dt):
        self.shipHeading.normalize()
        heading = self.shipHeading.angleDeg(Vec3(1,0,0))
        if self.shipHeading[1] < 0.0:
            heading = 360.0 - heading

        heading_rad = DegreeToRadiant * heading

        if self.keys["strafe left"]:
            heading += dt * TurnRate
        if self.keys["strafe right"]:
            heading -= dt * TurnRate
        self.ship.setH(heading)

        newVelocity = Vec3(0,0,0)
        if self.keys["accelerate"]:
            newVelocity = self.shipHeading * Acceleration * dt
            newVelocity += self.getVelocity(self.ship)

        if self.keys["break"]:
            # reverse direction
            newVelocity = self.shipHeading * Acceleration * dt * -1.0
            newVelocity += self.getVelocity(self.ship)
        
        if newVelocity.lengthSquared() > MaxVelocitySquared:
            newVelocity.normalize()
            newVelocity *= MaxVelocity

        self.setVelocity(self.ship, newVelocity)

        self.updatePosition(self.ship, dt)

    def positionCamera(self, task):
        # XXX: See gameLoop() comment
        return Task.cont
        position = self.ship.getPos()
        position[2] = CameraHeight
        self.camera.setPos(position)
        self.camera.setHpr(0, -90, 0)

        return Task.cont

    def gameLoop(self, task):
        dt = task.time - task.last
        task.last = task.time

        self.updateShip(dt)
        # XXX: Have to set the camera position here or it will be sluggish as hell
        position = self.ship.getPos()
        position[2] = CameraHeight
        self.camera.setPos(position)
        self.camera.setHpr(0, -90, 0)


        if self.keys["fire"] and task.time > task.nextBullet:
            self.fire(task.time, self.ship)
            task.nextBullet = task.time + BulletRepeat


        self.updateBullets(dt, task.time)

        if len(self.enemies) == 0:
            self.enemyCounterText.setText("You did it in %.1f seconds!" % self.neededTimeForLevel)
        else:
            self.neededTimeForLevel += globalClock.getDt()
            self.enemyCounterText.setText("Enemies: %i / %i" % (len(self.enemies), self.initialNumberOfEnemies))
        return Task.cont

    def mouseMovement(self, task):
        # Check to make sure the mouse is readable
        if base.mouseWatcherNode.hasMouse():
            # get the mouse position as a Vec2. The values for each axis are from -1 to
            # 1. The bottom-left is (-1,-1), the top right is (1,1)
            mpos = base.mouseWatcherNode.getMouse()

            self.shipHeading = Vec3(restrain(mpos.getX()), restrain(mpos.getY()), 0)

        return Task.cont


    def fire(self, time, ship):
        "spawn a new bullet for this ship with the selected weapon"
        position = ship.getPos()
        bullet = loadObject("plane", None, self.render, depth=PlayerHeight, color=Vec3(1, 0, 0), scale=0.1)
        bullet.setPos(position)
        bullet.setTwoSided(True)
        # Velocity is in relation to ship
        velocity = self.shipHeading * BulletSpeed
        self.setVelocity(bullet, velocity)
        self.setExpires(bullet, time+BulletLife)

        self.bullets.append(bullet)


    def updateBullets(self, dt, time):
        "Update all bullet positions and check for hits"
        newBullets = []
        wallScale = Vec3(1,1,1)

        for bullet in self.bullets:
            self.updatePosition(bullet, dt)

            if self.getExpires(bullet) > time:

                hitSomething = False
                for enemy in self.enemies:
                    if (bullet.getPos() - enemy.getPos()).lengthSquared() < (((bullet.getScale().getX() + enemy.getScale().getX()) * 0.5 ) ** 2):
                        self.setExpires(bullet, 0)
                        bullet.remove()
                        enemy.remove()
                        self.enemies.remove(enemy)
                        hitSomething = True
                        break

                if not hitSomething:
                    for wall in self.walls:
                        if (bullet.getPos() - wall).lengthSquared() < (((bullet.getScale().getX() + wallScale.getX()) * 0.5 ) ** 2):
                            self.setExpires(bullet, 0)
                            bullet.remove()
                            hitSomething = True
                            break

                if not hitSomething:
                    newBullets.append(bullet)
           
            else:
                bullet.remove()

        self.bullets = newBullets

    def loadLevel(self, path):
        f = open(os.path.normpath(path), "rt")
        lines = f.readlines()
        f.close()

        x, y = 0.0, 0.0
        tiles = {}

        tileColors = {}
        tileColors["f"] = (0.5, 0.5, 0.5)
        tileColors["X"] = (0.1, 0.1, 0.1)
        tileColors["1"] = (0.4, 0.05, 0.05)
        tileColors["2"] = (0.7, 0.05, 0.05)
        tileColors["void"] = (0.0, 0.0, 0.0)
        tileColors["S"] = (0.05, 0.8, 0.05)

        tileSizes = {"X":(1.0, 1.0, 1.0), "1":(0.3, 0.3, 1.0),  "2":(0.7, 0.7, 1.0), 
                    "void":(1.0, 1.0, 1.0), "S":(0.7, 0.7, 1.0), "f":(1.0, 1.0, 1.0)
                    }

        tileTextures = { "S":"ship" }

        for line in lines:
            tokens = line.split(";")
            x = 0.0
            for t in tokens:
                token = t.replace("\"", "").replace("\n", "")
                if token == "":
                    continue
                tiles[(int(x), int(y))] = token

                # TODO: argh, implicit filtering. 
                if token == "X" or token == "x":
                    tile = loadObject("plane", tileTextures.get(token, "Pointlight"), 
                                    self.levelRootNode, Point2(x, y), 
                                    color=tileColors[token], transparency=False)
                
                    tileCollider = tile.attachNewNode(CollisionNode("wall (%i, %i)" % (x,y)))
                    # make the sphere a little bit bigger to fake a box
                    tileCollider.node().addSolid(CollisionSphere(Point3(0,0,0), 0.6))
                    #tileCollider.show()
                    tileCollider.node().setFromCollideMask(BitMask32.allOff())
                    tileCollider.node().setIntoCollideMask(BitMask32(0x1))
                else:
                    token = "f"
                    tile = loadObject("plane", tileTextures.get(token, "Pointlight"), 
                                    self.levelRootNode, Point2(x, y), 
                                    color=tileColors[token], transparency=False)

                x += 1.0
            y += 1.0

        return (x,y), tiles

if __name__ == "__main__":

    loadPrcFileData("", "sync-video 1")
    loadPrcFileData("", "show-frame-rate-meter  #t")
    

    app = Game()
    run()

